#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 29/09/19, 11:39 PM
#
#
# Basic Objects

from SecretPlots.objects.shapes import *
from SecretPlots.utils import Log


class Data:
    """
    Class to hold and transform data

    Following are the data types

    0: Single Valued
    Only single not iterable number
    e.g. 1, 23, -8

    1: Points
    Iterable with exactly two length
    if sub-iterators, they should have same length
    e.g. [1,2] , [[1,1], [2,2]]

    2: Simple Categorical
    Iterator with length more than 2 with no sub-iterators
    e.g. [1,2,3,4]

    3: Matrix
    Iterator with sub-iterators with same length (and are not points)
    e.g. [[1,1,1], [2,2,2]]

    4: Complex Categorical
    Iterator with sub-iterators with uneven length
    e.g. [[1,1], [2,2,2], [3,3,3]], [[1,2,3,4]]

    Currently more than 2 Dimensions are not supported. So you should
    convert multidimensional data into 2D to fit into this framework

    """
    SINGLE_VALUED = 0
    POINTS = 1
    SIMPLE_CATEGORICAL = 2
    MATRIX = 3
    COMPLEX_CATEGORICAL = 4

    def __init__(self, data, log: Log):
        self._raw_data = data
        self._log = log
        self._type = None
        self._value = None
        self._positions = None
        self.threshold = None

    @property
    def max(self):
        return np.nanmax(self.value)

    @property
    def min(self):
        return np.nanmin(self.value)

    @property
    def value(self):
        """
        Flatten values of the entire dataset

        Argument 'dtype=np.float' is required for replacement of None
        """
        if self._value is None:
            if self.type != Data.COMPLEX_CATEGORICAL:
                self._value = np.asarray(self._raw_data,
                                         dtype=np.float).flatten()
            else:
                self._value = []
                for k in self._raw_data:
                    try:
                        self._value.extend(np.asarray(k))
                    except TypeError:
                        self._value.append(k)
                self._value = np.asarray(self._value, dtype=np.float)
        return self._value

    @property
    def positions(self):
        """
        Returns flatten array of positions of the (row, column) in given
        dataset
        """
        if self._positions is None:
            if self.type == Data.SINGLE_VALUED:
                self._positions = [(0, 0)]
            elif self.type == Data.POINTS:
                self._assign_point_locations()
            elif self.type == Data.SIMPLE_CATEGORICAL:
                self._positions = [(0, x) for x in range(len(self._raw_data))]
            elif self.type == Data.MATRIX:
                self._positions = []
                row, cols = np.asarray(self._raw_data).shape
                for r in range(row):
                    for c in range(cols):
                        self._positions.append((r, c))
            elif self.type == Data.COMPLEX_CATEGORICAL:
                self._assign_complex_locations()

            self._log.info("Locations assigned to the data points")

        return self._positions

    @property
    def type(self) -> int:
        if self._type is None:
            self._assign_type()
        return self._type

    @property
    def type_name(self) -> str:
        if self.type == Data.SINGLE_VALUED:
            return "Single Valued"
        elif self.type == Data.POINTS:
            return "Points"
        elif self.type == Data.SIMPLE_CATEGORICAL:
            return "Simple Categorical"
        elif self.type == Data.COMPLEX_CATEGORICAL:
            return "Complex Categorical"
        elif self.type == Data.MATRIX:
            return "Matrix"
        else:
            self._log.error("No such category found {}".format(self.type))

    @property
    def raw_data(self):
        return self._raw_data

    @property
    def is_single_point(self) -> bool:
        if self.type != Data.POINTS:
            return False
        else:
            try:
                iter(self.raw_data[0])
                return False
            except TypeError:
                return True

    def _assign_point_locations(self):
        self._positions = []
        for i, x in enumerate(self._raw_data):
            try:
                iter(x)
                for k in range(len(x)):
                    self.positions.append((i, k))
            except TypeError:
                self.positions.append((0, i))

    def _assign_complex_locations(self):
        self._positions = []
        for i, x in enumerate(self._raw_data):
            try:
                for j, k in enumerate(x):
                    self._positions.append((i, j))
            except TypeError:
                self._log.warn("Non-iterable element ({}) found in the "
                               "Complex Categorical dataset".format(x))
                self._positions.append((i, 0))

    def _assign_type(self):
        try:
            iter(self._raw_data)
            try:
                self._check_multi_dimension()
            except TypeError:
                self._log.error("Something went wrong in calculating "
                                " data dimensions. Please check input data.")
        except TypeError:
            # Single Valued
            self._type = Data.SINGLE_VALUED
            self._log.info("Data type assigned as Single Valued ({})".format(
                Data.SINGLE_VALUED))

    def _check_multi_dimension(self):
        temp = [np.asarray(m) for m in self._raw_data]
        temp_len = [m.size for m in temp]
        if len(temp) == 2 and temp_len.count(temp_len[0]) == len(temp_len):
            # Points
            self._type = Data.POINTS
        else:
            if temp_len.count(temp_len[0]) == len(temp_len) and len(
                    temp_len) != 1:

                if temp_len[0] == 1:
                    self._type = Data.SIMPLE_CATEGORICAL
                else:
                    self._type = Data.MATRIX
            else:
                self._type = Data.COMPLEX_CATEGORICAL

        self._log.info("Data type assigned as {} ({})".format(self.type,
                                                              self.type_name))


class Element:
    def __init__(self, log):
        self._log = log
        self.is_point = False
        self.width = 1
        self.height = 1
        self.value = None
        self.rotation = 0
        self._options = None
        self._shape = "r"

    @property
    def shape(self):
        if self._shape in ["r", "rectangle", "rect"]:
            return Rectangle
        elif self._shape in ["t", "triangle", "tri"]:
            return Triangle
        elif self._shape in ["c", "circle", "cir"]:
            return Circle
        else:
            self._log.error("Shape {} not found".format(self._shape))

    @shape.setter
    def shape(self, value: str):
        self._shape = value.strip().lower()

    @property
    def options(self):
        if self._options is None:
            self._options = {}
        return self._options

    def add_options(self, **kwargs):
        self._options = {**self.options, **kwargs}

    def get(self, x, y):
        return self.shape(x, y, self.width, self.height, self.rotation,
                          **self.options)


class Axis:
    def __init__(self, name: str, index: int, log: Log):
        self.name = name
        self.index = index
        self._log = log
        self.show_ticks = True
        self._ticks = None
        self._tick_labels = None
        self._ticklabels_options = None
        self._tick_directions = None
        self._tick_options = None
        self._gap = None
        self._gap_options = None
        self._user_defined_gap = False
        self.show_midlines = False
        self.show_edgelines = False
        self._midlines_options = None
        self._midlines = None
        self._edgelines_options = None
        self._edgelines = None
        self._label = None
        self._label_options = None

        self.scale = None
        self.limit = (0, 1)
        self.padding_start = None
        self.padding_end = None
        self.is_inverted = False
        self._missing_region = []

        self._log.info("{} axis instance is generated".format(name))

    @property
    def ticks(self) -> list:
        if self._ticks is None:
            self._ticks = []
        return self._ticks

    @ticks.setter
    def ticks(self, values):
        self._ticks = values

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def missing_regions(self):
        return self._missing_region

    @property
    def midlines(self):
        return self._midlines

    @property
    def edgelines(self):
        if self._edgelines is None:
            self._edgelines = []
        return self._edgelines

    @midlines.setter
    def midlines(self, values):
        self._midlines = values

    @edgelines.setter
    def edgelines(self, values):
        self._edgelines = values

    @property
    def label_options(self):
        if self._label_options is None:
            self._label_options = {}
        return self._label_options

    @property
    def midlines_options(self):
        if self._midlines_options is None:
            self._midlines_options = {}
        return self._midlines_options

    @property
    def edgelines_options(self):
        if self._edgelines_options is None:
            self._edgelines_options = {}
        return self._edgelines_options

    @property
    def ticklabels_options(self):
        if self._ticklabels_options is None:
            self._ticklabels_options = {}
        return self._ticklabels_options

    @property
    def tick_options(self):
        if self._tick_options is None:
            self._tick_options = {}
        return self._tick_options

    @property
    def tick_labels(self) -> list:
        if self._tick_labels is None:
            self._tick_labels = []
        return self._tick_labels

    @tick_labels.setter
    def tick_labels(self, values):
        self._tick_labels = values

    @property
    def gap(self) -> float:
        if self._gap is None:
            self._gap = 0
        return self._gap

    @gap.setter
    def gap(self, value: float):
        self._user_defined_gap = True
        self._gap = value

    @property
    def is_user_defined_gap(self):
        return self._user_defined_gap

    def add_midlines_options(self, **kwargs):
        self._midlines_options = {**self.midlines_options, **kwargs}

    def add_edgelines_options(self, **kwargs):
        self._edgelines_options = {**self.edgelines_options, **kwargs}

    def add_ticklabels_options(self, **kwargs):
        self._ticklabels_options = {**self.ticklabels_options, **kwargs}

    def add_tick_options(self, **kwargs):
        self._tick_options = {**self.tick_options, **kwargs}

    def add_label_options(self, **kwargs):
        self._label_options = {**self.label_options, **kwargs}

    def add_missing_region(self, p1, length):
        self._missing_region.append((p1, p1 + length))

    def make_ticks(self, values):
        if self._ticks is not None:
            self._log.info("Ticks for {} have already been set by "
                           "user".format(self.name))
            return
        self._ticks = values
        self._log.info("Ticks for {} automatically set".format(self.name))

    def make_labels(self):
        if self._tick_labels is not None:
            self._log.info("Tick labels for {} are already defined".format(
                self.name))
            return
        self._tick_labels = ["{}".format(x) for x in range(len(self.ticks))]
        self._log.info("{} tick_labels automatically generated".format(
            self.name))

    def make_midlines(self):
        if self.midlines is not None:
            self._log.info("Generation of midlines ignored because they are "
                           "already generated")
            return

        self._midlines = []
        if len(self.ticks) == 1:
            self._log.warn("{} midlines could not be generated because only "
                           "1 bar is present".format(self.name))
            return

        for i, x in enumerate(self.ticks[:-1]):
            self._midlines.append((x + self.ticks[i + 1]) / 2)
        self._log.info("{} midlines automatically generated".format(self.name))


def run():
    pass



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#
#
#  All Objects


__author__ = "Rohit Suratekar"
__email__ = "rohitsuratekar@gmail.com"
__license__ = "MIT License"
__status__ = "Prototype"

from SecretPlots.graphs.categorical import (BarPlot,
                                            ColorPlot,
                                            BooleanPlot,
                                            BarGroupedPlot)

from SecretPlots.network.graphs import NetworkPlot, Space



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 11/15/19, 11:09 AM
#
#
# Path finder functions related to the our network

from typing import List, Dict, Union

import numpy as np


# np.set_printoptions(linewidth=300)


class Point:
    def __init__(self, index: int):
        self.index = index
        self.row = None
        self.column = None
        self.mark = "."
        self.value = np.inf
        self.original_value = None
        self.previous_node = None

    @property
    def pos(self):
        return self.row, self.column

    def __repr__(self):
        return f"{self.mark}"
        # return f"({self.value}, {self.previous_node})"

    def update(self, value, node):
        if value < self.value:
            self.value = value
            self.previous_node = node


class PathFinder:
    def __init__(self, graph):
        self._raw_data = np.asanyarray(graph)
        self._graph = None
        self.path_mark = "o"
        self.empty_mark = "."
        self.block_mark = "|"
        self.end_mark = "X"
        self.start_mark = "o"
        self.block_item = None
        self.block_penalty = np.inf
        self.show_blocks = False

    @property
    def graph(self) -> np.ndarray:
        if self._graph is None:
            self._generate_nodes()
        return self._graph

    @property
    def columns(self):
        return self.graph.shape[1]

    @property
    def rows(self):
        return self.graph.shape[0]

    def _generate_nodes(self):
        data = self._raw_data.flatten()
        new_data = []
        for i in range(len(data)):
            node = Point(i)
            node.row = int(i / self._raw_data.shape[1])
            node.column = i % self._raw_data.shape[1]
            node.original_value = data[i]
            node.mark = self.empty_mark
            if self.show_blocks and node.original_value == self.block_item:
                node.mark = self.block_mark
            new_data.append(node)

        data = new_data
        data = np.array(data, dtype=Point).reshape(self._raw_data.shape)
        self._graph = data

    def get_surrounding(self, row, col) -> List[Point]:
        def _get_item(r, c):
            if r < 0 or c < 0 or r >= self.rows or c >= self.columns:
                return None
            else:
                return self.graph[r, c]

        data = [
            _get_item(row, col + 1),  # Right
            _get_item(row + 1, col),  # Bottom
            _get_item(row, col - 1),  # Left
            _get_item(row - 1, col)  # Top

        ]
        return [x for x in data if x is not None]

    def dijkstra_distance(self, start_node: Point, end_node: Point):
        if end_node.original_value == self.block_item:
            return self.block_penalty
        return start_node.value + 1

    @staticmethod
    def cartesian_distance(node1: Point, node2: Point):
        y1, x1 = node1.pos
        y2, x2 = node2.pos
        return np.sqrt(np.square(x1 - x2) + np.square(y1 - y2))

    def a_star_distance(self,
                        start_node: Point,
                        end_node: Point,
                        final_node: Point):

        if end_node.original_value == self.block_item:
            return self.block_penalty
        return start_node.value + self.cartesian_distance(end_node, final_node)

    def _generate_distance_matrix(self,
                                  start_index: int,
                                  end_index: int, *,
                                  use_dijkstra: bool):

        start_row, start_col = self.split_index(start_index, self.columns)
        end_row, end_col = self.split_index(end_index, self.columns)
        self.graph[start_row, start_col].value = 0
        unvisited = [x for x in self.graph.flatten()]
        while len(unvisited) >= 1:
            unvisited = sorted(unvisited, key=lambda x: x.value)
            current = unvisited[0]
            row, col = current.pos
            for a in self.get_surrounding(row, col):
                if use_dijkstra:
                    a.update(self.dijkstra_distance(current, a), current.index)
                else:
                    a.update(
                        self.a_star_distance(current,
                                             a,
                                             self.graph[end_row, end_col]),
                        current.index)
            unvisited.remove(current)

    @staticmethod
    def split_index(index, width):
        row = int(index / width)
        col = index % width
        return row, col

    def _trace_path(self, start, end) -> list:
        path = []
        row, col = self.split_index(end, self.columns)
        item = self.graph[row, col]  # type:Point
        path.append(item.index)
        while item != start:

            item.mark = self.path_mark
            if item.index == end:
                item.mark = self.end_mark
            elif item.index == start:
                item.mark = self.start_mark
            if item.previous_node is None:
                break
            nr, nc = self.split_index(item.previous_node, self.columns)
            item = self.graph[nr, nc]
            path.append(item.index)

        path = list(reversed(path))
        return path

    def find_path(self, start, end, *, use_dijkstra=False):
        self._generate_distance_matrix(start, end, use_dijkstra=use_dijkstra)
        self._trace_path(start, end)
        print(self.graph)
        # Reset Data
        self._generate_nodes()

    def boolean_path(self, start, end, *,
                     use_dijkstra=False,
                     exclude_start=False,
                     exclude_end=False) -> np.ndarray:

        self._generate_distance_matrix(start, end, use_dijkstra=use_dijkstra)
        self._trace_path(start, end)
        data = self.graph.flatten()
        for i in range(len(data)):
            if data[i].mark in [self.path_mark, self.end_mark]:
                data[i] = 1
            else:
                data[i] = 0
        if exclude_start:
            data[start] = 0
        if exclude_end:
            data[end] = 0
        data = np.asanyarray(data).reshape(self.graph.shape)
        return data

    def path_index(self, start, end, *,
                   use_dijkstra=False,
                   exclude_start=False,
                   exclude_end=False) -> list:

        self._generate_distance_matrix(start, end, use_dijkstra=use_dijkstra)
        return self._trace_path(start, end)


class MatItem:

    def __init__(self, height, width):
        self.x = None
        self.y = None
        self.height = height
        self.width = width
        self.name = "N/A"
        self.column = None
        self.row = None

    def assign(self, index: int, width: int):
        row = int(index / width)
        col = index % width
        self.x = col * self.width
        self.y = row * self.height
        self.column = col
        self.row = row

    @property
    def is_gap(self):
        raise NotImplementedError()


class Edge:
    def __init__(self, start, end, link_type):
        self.start = start
        self.end = end


class Node(MatItem):
    @property
    def is_gap(self):
        return False

    def __init__(self, name, height, width):
        super().__init__(height, width)
        self.name = name
        self.links = 0
        self.paths = {}
        self.color = None

    def __repr__(self):
        return f"({self.name} {self.links})"

    def add_paths(self, node_name, indexes):
        self.paths[node_name] = indexes


class Gap(MatItem):
    @property
    def is_gap(self):
        return True

    def __init__(self, height, width):
        super().__init__(height, width)
        self.lines = []
        self.v_slots = 1
        self.h_slots = 1

    def add_line(self, node_name: str):
        if node_name not in self.lines:
            self.lines.append(node_name)

    def __repr__(self):
        if len(self.lines) == 0:
            return f"(   )"
        else:
            return f"( {''.join(self.lines)} )"


class Space:
    def __init__(self, data, height, width):
        self._raw_data = data
        self._nodes = None
        self._max_cols = None
        self._matrix = None
        self.node_gap = 1
        self.use_dijkstra = True
        self.node_height = height
        self.node_width = width
        self._assignment_done = False

    @property
    def nodes(self) -> Dict[str, Node]:
        if self._nodes is None:
            nodes = {}
            # Create empty nodes
            for d in self._raw_data:
                nodes[d[0]] = Node(d[0], height=self.node_height,
                                   width=self.node_width)
                nodes[d[1]] = Node(d[1], height=self.node_height,
                                   width=self.node_width)

            # Add details
            for d in self._raw_data:
                nodes[d[0]].links += 1
                nodes[d[1]].links += 1

            self._nodes = nodes
        return self._nodes

    @property
    def max_cols(self) -> int:
        if self._max_cols is None:
            self._max_cols = int(np.sqrt(len(self.nodes)))
        return self._max_cols

    @max_cols.setter
    def max_cols(self, value: int):
        if not isinstance(value, int):
            raise ValueError(f"Column size can not be {type(value)}")
        self._max_cols = value

    @property
    def matrix(self) -> np.ndarray:
        if self._matrix is None:
            self._place_nodes()
        return self._matrix

    def _place_nodes(self):

        mat = [x for x in self.nodes.values()]
        forward_array = []
        reverse_array = []
        swap_array = True
        while len(mat) > 0:
            mat = sorted(mat, key=lambda x: x.links, reverse=True)
            if swap_array:
                forward_array.append(mat[0])
            else:
                reverse_array.append(mat[0])
            mat.pop(0)
            swap_array = not swap_array
        mat = forward_array
        mat.extend(reversed(reverse_array))

        while len(mat) % self.max_cols != 0:
            mat.insert(-1, None)

        mat = np.array(mat).reshape((-1, self.max_cols))

        self._matrix = mat

    def _add_space(self):
        ng = self.node_gap + 1
        rows = self.matrix.shape[0] * ng + (ng - 1)
        cols = self.matrix.shape[1] * ng + (ng - 1)
        a = [None] * rows * cols
        a = np.array(a).reshape((rows, cols))
        a[(ng - 1)::ng, (ng - 1)::ng] = self.matrix
        k = [Gap(height=self.node_height,
                 width=self.node_width) if x is None else x
             for x in a.flatten()]
        k = np.array(k).reshape(a.shape)
        self._matrix = k

    @property
    def boolean_matrix(self) -> np.ndarray:
        x = [0 if m.is_gap else None for m in self.matrix.flatten()]
        x = np.array(x).reshape(self.matrix.shape)
        return x

    def _get_node_coord(self, node: Node) -> tuple:
        k = list(self.matrix.flatten())
        k = k.index(node)
        return self._convert_idx(k)

    def _get_node_index(self, node: Node, is_output: bool):
        row, col = self._get_node_coord(node)
        if is_output:
            row -= 1  # TODO: Check properly
        else:
            row += 1
        return row * self.matrix.shape[1] + col

    def _get_item_at(self, index: int) -> Union[Gap, Node]:
        row, col = self._convert_idx(index)
        return self.matrix[row, col]

    def _convert_idx(self, idx) -> tuple:
        row = int(idx / self.matrix.shape[1])
        col = idx % self.matrix.shape[1]
        return row, col

    def _assign_edges(self):
        for d in self._raw_data:
            p = PathFinder(self.boolean_matrix)
            start_node = self._get_node_index(self.nodes[d[0]], True)
            end_node = self._get_node_index(self.nodes[d[1]], False)
            idx = []
            for i in p.path_index(start_node, end_node,
                                  use_dijkstra=self.use_dijkstra):
                r, c = self._convert_idx(i)
                idx.append((i, r, c))

            self.nodes[d[0]].add_paths(d[1], idx)

    def get(self) -> np.ndarray:
        if not self._assignment_done:
            self._add_space()
            self._assign_edges()
            self._assignment_done = True
        return self.matrix


def run():
    data = [
        ["a", "b", 1],
        ["a", "c", 1],
        ["a", "d", 1],
        ["c", "d", 1],
        ["b", "d", 1],
    ]



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 11/19/19, 12:13 PM
#
#
# Network plots

from collections import defaultdict

import matplotlib.pyplot as plt
from SecretColors import Palette
from SecretColors.utils import text_color
from matplotlib.lines import Line2D

from SecretPlots.constants.network import *
from SecretPlots.network.pathfinder import *
from SecretPlots.objects.shapes import Rectangle
from SecretPlots.utils import Log


class NetworkPlot:
    def __init__(self, data, fig: plt.Figure = None, log: Log = None):
        if log is None:
            log = Log()
        if fig is None:
            fig = plt.figure()

        self._log = log
        self.data = data
        self.palette = Palette()
        self._space = None
        self.fig = fig  # type: plt.Figure

        # Options
        self.use_dijkstra = True
        self.colors = None
        self.colors_mapping = None
        self.node_width = 1
        self.node_height = 1
        self.node_gap = 1
        self.max_columns = None
        self.aspect_ratio = None
        self.font_size = 13

        self._ax = None
        self._text_options = None
        self._line_options = None
        self._fig_drawn = False
        self._all_nodes = None

        self._log.info("Network Plot is initialized")

    @property
    def ax(self) -> plt.Axes:
        if self._ax is None:
            self._ax = self.fig.subplots()
        return self._ax

    @property
    def space(self) -> Space:
        if self._space is None:
            self._space = Space(self.data,
                                height=self.node_height,
                                width=self.node_width)
            self._space.use_dijkstra = self.use_dijkstra
            self.node_gap = self.node_gap
            if self.max_columns is not None:
                self._space.max_cols = self.max_columns
        return self._space

    def _color_cycle(self):
        while True:
            if self.colors is None:
                for p in self.palette:
                    yield p
            else:
                if type(self.colors) == str:
                    self.colors = [self.colors]
                for c in self.colors:
                    yield c

    @property
    def text_options(self) -> dict:
        if self._text_options is None:
            self._text_options = {
                "ha": "center",
                "va": "center",
                "fontsize": self.font_size
            }
        return self._text_options

    @property
    def line_options(self) -> dict:
        if self._line_options is None:
            self._line_options = {}
        return self._line_options

    @property
    def all_nodes(self) -> Dict[str, Node]:
        if self._all_nodes is None:
            self._draw_elements()
        return self._all_nodes

    def _add_text(self, r: Rectangle, text: str):
        self.ax.text(
            r.x + r.width / 2,
            r.y + r.height / 2,
            text,
            color=text_color(r.color),
            **self.text_options
        )

    def add_text_options(self, **kwargs):
        self._text_options = {**self.text_options, **kwargs}

    def add_line_options(self, **kwargs):
        self._line_options = {**self.line_options, **kwargs}

    @staticmethod
    def _get_direction(p1: tuple, p2: tuple, p3: tuple) -> tuple:
        def _check(x1, x2, y1, y2):
            if x1 == x2:
                if y1 < y2:
                    return BOTTOM
                else:
                    return TOP
            else:
                if x1 < x2:
                    return LEFT
                else:
                    return RIGHT

        input_direction = _check(p1[0], p2[0], p1[1], p2[1])
        output_direction = _check(p3[0], p2[0], p3[1], p2[1])

        return input_direction, output_direction

    def _check_line_direction(self,
                              gap: Gap, pg: Union[Gap, None],
                              ng: Union[Gap, None]) -> tuple:
        x2, y2 = gap.column, gap.row
        if pg is None:
            x1, y1 = gap.column, gap.row + 1
        else:
            x1, y1 = pg.column, pg.row
        if ng is None:
            x3, y3 = gap.column, gap.row - 1
        else:
            x3, y3 = ng.column, ng.row

        return self._get_direction((x1, y1), (x2, y2), (x3, y3))

    def _add_line(self,
                  gap: Gap,
                  pg: Gap,
                  ng: Gap,
                  row_slots: Dict,
                  col_slots: Dict,
                  line: str,
                  color: str):

        arrow_width = 0.05
        arrow_height = 0.1
        rf = row_slots[gap.row]
        f = 1 / (len(rf) + 1)
        col_factor = f + f * rf.index(line)

        cf = col_slots[gap.column]
        e = 1 / (len(cf) + 1)
        row_factor = e + e * cf.index(line)

        def _points(direction):
            if direction == LEFT:
                x.append(gap.x)
                y.append(gap.y + gap.height * col_factor)
            elif direction == RIGHT:
                x.append(gap.x + gap.width)
                y.append(gap.y + gap.height * col_factor)
            elif direction == TOP:
                x.append(gap.x + gap.width * row_factor)
                y.append(gap.y + gap.height)
            else:
                x.append(gap.x + gap.width * row_factor)
                y.append(gap.y)

        in_dir, out_dir = self._check_line_direction(gap, pg, ng)
        x = []
        y = []
        if pg is None and ng is None:
            x.extend(
                [gap.x + gap.width * row_factor,
                 gap.x + gap.width * row_factor]
            )
            y.extend(
                [gap.y + gap.height, gap.y]
            )
        elif pg is None:
            x.extend(
                [gap.x + gap.width / 2, gap.x + gap.width / 2])
            y.extend([gap.y + gap.height,
                      gap.y + gap.height * col_factor,
                      gap.y + gap.height * col_factor])
            if out_dir == LEFT:
                x.append(gap.x)
            else:
                x.append(gap.x + gap.width)

        else:

            _points(in_dir)

            x.append(gap.x + gap.width * row_factor)
            y.append(gap.y + gap.height * col_factor)

            if ng is None:
                x.append(gap.x + gap.width * row_factor)
                y.append(gap.y)

            else:
                _points(out_dir)

        opts = {"color": color}
        opts = {**opts, **self.line_options}
        line = Line2D(x, y, **opts)
        self.ax.add_line(line)

        if ng is None:
            # This is last block, add arrow
            arrow_x = [
                gap.x + gap.width * row_factor - arrow_width,
                gap.x + gap.width * row_factor,
                gap.x + gap.width * row_factor + arrow_width
            ]
            arrow_y = [
                gap.y + arrow_height,
                gap.y,
                gap.y + arrow_height
            ]

            arrow_line = Line2D(arrow_x, arrow_y, **opts)
            self.ax.add_line(arrow_line)

    @staticmethod
    def _arrange_lines(data: np.ndarray):
        # NOTE: This function will modify the original data
        # Assign gaps
        row_slots = defaultdict(list)
        col_slots = defaultdict(list)
        for item in data.flatten():
            if not item.is_gap:
                for val in item.paths:
                    for _, r, c in item.paths[val]:
                        if item.name not in row_slots[r]:
                            row_slots[r].append(item.name)
                        if item.name not in col_slots[c]:
                            col_slots[c].append(item.name)
                        data[r, c].add_line(item.name)

        # Assign slots
        for item in data.flatten():  # type: Union[Gap, Node]
            if item.is_gap:
                item.h_slots = len(row_slots[item.row])
                item.v_slots = len(col_slots[item.column])

        return row_slots, col_slots

    def _draw_lines(self):
        # Draw Lines
        row_slots, col_slots = self._arrange_lines(self.space.get())

        for n in self.all_nodes.values():
            for b in n.paths:
                vals = [self.space.matrix[x[1], x[2]] for x in n.paths[b]]
                vals.insert(0, None)
                vals.append(None)
                for i in range(len(vals))[1:-1]:
                    self._add_line(vals[i],
                                   vals[i - 1],
                                   vals[i + 1],
                                   row_slots,
                                   col_slots,
                                   n.name,
                                   n.color)

    def _get_color(self, node_name, cycle) -> str:
        if self.colors_mapping is None:
            return next(cycle)
        else:
            if node_name in self.colors_mapping.keys():
                return self.colors_mapping[node_name]
            else:
                return next(cycle)

    def _draw_elements(self):
        if self._all_nodes is not None:
            self._log.warn("Elements have been already drawn")
            return

        palette = self._color_cycle()
        s = [x for x in
             self.space.get().flatten()]  # type: List[Union[MatItem, Node]]
        # Add points to the node
        for i, m in enumerate(s):
            m.assign(i, self.space.matrix.shape[1])

        all_nodes = {}

        all_x = []
        all_y = []
        # Draw nodes
        for m in s:
            all_x.extend([m.x, m.x + m.width])
            all_y.extend([m.y, m.y + m.height])
            if not m.is_gap:
                r = Rectangle(m.x, m.y, height=m.height, width=m.width,
                              color=self._get_color(m.name, palette))
                m.color = r.color
                self.ax.add_patch(r.get())
                self._add_text(r, m.name)
                all_nodes[m.name] = m

        self._all_nodes = all_nodes
        self.ax.set_xlim(min(all_x), max(all_x))
        self.ax.set_ylim(min(all_y), max(all_y))
        self._log.info("All nodes arranged in the space")

    def draw(self):
        # Sanity check to avoid redrawing
        if self._fig_drawn:
            self._log.warn("Figure is already drawn.")
            return
        else:
            self._fig_drawn = True

        self._draw_elements()
        self._draw_lines()
        self.ax.set_axis_off()
        if self.aspect_ratio is not None:
            self.ax.set_aspect(self.aspect_ratio)

    def show(self, tight=False):
        self.draw()
        if tight:
            plt.tight_layout()
        plt.show()

    def save(self, filename, **kwargs):
        self.draw()
        plt.savefig(filename, **kwargs)


def run():
    data = [
        ["a", "b", 1],
        ["a", "c", 1],
        ["c", "d", 1],
        ["c", "a", 1],
        ["c", "a1", 1],
        ["c", "a2", 1],
        ["c", "a2", 1],
        ["c", "ae2", 1],
        ["c", "ae32", 1],
        ["c", "ta", 1],
        ["b", "ta", 1],
    ]

    n = NetworkPlot(data)
    n.show()



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 19/09/19, 10:41 AM
#
#
# All Shapes will be accessed from this script


import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
from SecretColors import Palette
from matplotlib.path import Path

palette = Palette()


class Rectangle:
    def __init__(self, x, y, width, height,
                 rotation: float = 0.0,
                 align: str = "center",
                 **kwargs):
        self._x = x
        self._y = y
        self._height = height
        self._width = width
        self._rotation = rotation
        self._align = align
        self._options = kwargs

    @property
    def align(self):
        return self._align

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    @property
    def rotation(self):
        return self._rotation

    @property
    def options(self):
        if "color" not in self._options.keys():
            self._options["color"] = palette.blue()
        return self._options

    def get(self):
        return patches.Rectangle((self.x, self.y),
                                 self.width,
                                 self.height,
                                 self.rotation, **self.options)

    @property
    def color(self):
        return self.options["color"]


class Cuts:
    def __init__(self, obj,
                 no_of_cuts: int = 3,
                 cut_direction: str = None,
                 cut_depth: float = 1,
                 position: str = "top",
                 **kwargs):
        self._object = obj
        self.no_of_cuts = no_of_cuts
        self._cut_direction = cut_direction
        self.cut_depth = cut_depth
        self.position = position
        self._options = kwargs

    @property
    def object(self):
        return self._object

    @property
    def options(self):
        return self._options

    @staticmethod
    def _real_quadrant(angle):
        return abs(angle % 360)

    @property
    def cut_direction(self):
        if self._cut_direction is None:
            if self.position == "top":
                if 0 < self._real_quadrant(self.object.rotation) <= 180:
                    self._cut_direction = "in"
                else:
                    self._cut_direction = "out"
            elif self.position == "right":
                if 0 <= self._real_quadrant(self.object.rotation) < 180:
                    self._cut_direction = "out"
                else:
                    self._cut_direction = "in"
            elif self.position == "bottom":
                if 0 < self._real_quadrant(self.object.rotation) <= 180:
                    self._cut_direction = "out"
                else:
                    self._cut_direction = "in"
            elif self.position == "left":
                if 0 <= self._real_quadrant(self.object.rotation) < 180:
                    self._cut_direction = "in"
                else:
                    self._cut_direction = "out"
            else:
                raise Exception("Unknown cut position {}".format(
                    self.position))
        return self._cut_direction

    @property
    def start(self):
        if self.position in ["bottom", "left"]:
            return self.object.get().get_verts()[0]
        elif self.position == "right":
            return self.object.get().get_verts()[1]
        elif self.position == "top":
            return self.object.get().get_verts()[3]
        else:
            raise Exception("Unknown cut position {}".format(self.position))

    @property
    def end(self):
        if self.position == "bottom":
            return self.object.get().get_verts()[1]
        elif self.position == "left":
            return self.object.get().get_verts()[3]
        elif self.position in ["top", "right"]:
            return self.object.get().get_verts()[2]
        else:
            raise Exception("Unknown cut position {}".format(self.position))

    @staticmethod
    def simple_connect(vert: list):
        code = [Path.MOVETO]
        code.extend([Path.LINETO] * (len(vert) - 2))
        code.append(Path.CLOSEPOLY)
        return code

    def _calculate_points(self):
        p = get_points_between(self.start,
                               self.end,
                               self.no_of_cuts,
                               self.cut_depth,
                               self.cut_direction)
        p.append(self.start)
        self._options = self.object.options
        return Path(p, self.simple_connect(p))

    def get(self):
        p = self._calculate_points()
        return patches.PathPatch(p, **self.options)


def get_points_between(start, end, cuts, depth, direction):
    x1, y1 = start
    x2, y2 = end

    x = np.linspace(x1, x2, cuts + 1)
    y = np.linspace(y1, y2, cuts + 1)

    points = []

    for i in range(len(x))[:-1]:
        points.append((x[i], y[i]))
        tx, ty = _get_cut_point((x[i], y[i]), (x[i + 1], y[i + 1]),
                                depth,
                                direction=direction)
        points.append((tx, ty))

    points.append((x[-1], y[-1]))
    return points


def _get_cut_point(p1, p2, height, direction="out"):
    x1, y1 = p1
    x2, y2 = p2

    factor = 1
    if direction == "in":
        factor = -1

    if x2 - x1 == 0:
        return x1 + factor * height, (y1 + y2) / 2

    # Find slope
    slope = (y2 - y1) / (x2 - x1)

    if slope == 0:
        return (x1 + x2) / 2, y1 + factor * height

    # Perpendicular line slope
    slope = -1 / slope

    # Mid point will be on that line
    mid_x = (x1 + x2) / 2
    mid_y = (y1 + y2) / 2

    # Find C intercept of this line
    c = mid_y - (slope * mid_x)

    # After solving equation for slop and circle
    x = mid_x + height * factor / (np.sqrt(np.square(slope) + 1))
    y = x * slope + c

    return x, y


class Triangle:
    def __init__(self, x, y, width, height, rotation=0, **kwargs):
        self._x = x
        self._y = y
        self._height = height
        self._width = width
        self._rotation = rotation
        self._options = kwargs

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    @property
    def rotation(self):
        return self._rotation

    @property
    def options(self):
        if "color" not in self._options.keys():
            self._options["color"] = palette.blue()
        return self._options

    def _get_rect(self):
        return patches.Rectangle((self.x, self.y), self.width, self.height,
                                 self.rotation)

    def get(self):
        p = self._get_rect()
        vert = p.get_verts()
        t1 = vert[2]
        t2 = vert[3]
        vert = [x for x in vert[:2]]
        vert.append(np.asarray([(t1[0] + t2[0]) / 2, (t1[1] + t2[1]) / 2]))
        vert.append(vert[0])
        return patches.Polygon(vert, **self.options)


class Circle:
    def __init__(self, x, y, width, height, rotation=0, **kwargs):
        self._x = x
        self._y = y
        self._height = height
        self._width = width
        self._rotation = rotation
        self._options = kwargs

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    @property
    def rotation(self):
        return self._rotation

    @property
    def options(self):
        if "color" not in self._options.keys():
            self._options["color"] = palette.blue()
        return self._options

    def _get_rect(self):
        return patches.Rectangle((self.x, self.y), self.width, self.height,
                                 self.rotation)

    @staticmethod
    def _get_diagonal_mid(points):
        p1, p2, p3, p4 = points

        m1 = (p3[1] - p1[1]) / (p3[0] - p1[0])
        m2 = (p4[1] - p2[1]) / (p4[0] - p2[0])

        c1 = p3[1] - m1 * p3[0]
        c2 = p4[1] - m2 * p4[0]

        x = (c2 - c1) / (m1 - m2)
        y = m1 * x + c1

        return x, y

    def get(self):
        verts = self._get_rect().get_verts()
        x, y = self._get_diagonal_mid(verts[:-1])
        return patches.Ellipse((x, y),
                               self.width,
                               self.height,
                               self.rotation,
                               **self.options)


def run():
    fig, ax = plt.subplots()
    p = Circle(0, 0, 1, 1)
    p2 = Rectangle(0, 0, 1, 1, fill=False)
    ax.add_patch(p.get())
    ax.add_patch(p2.get())
    ax.set_xlim(-3, 3)
    ax.set_ylim(-3, 3)
    ax.axhline(0)
    ax.axvline(0)

    plt.show()



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 28/09/19, 1:06 AM
#
# All categorical plots

from SecretPlots.assemblers import (Assembler,
                                    BarAssembler,
                                    BooleanAssembler,
                                    BarGroupedAssembler,
                                    ColorMapAssembler)
from SecretPlots.graphs._graphs import SecretPlot


class BarPlot(SecretPlot):

    @property
    def main_assembler(self) -> Assembler:
        return BarAssembler(self.fig, self._log)


class BarGroupedPlot(SecretPlot):

    @property
    def main_assembler(self) -> Assembler:
        return BarGroupedAssembler(self.fig, self._log)


class ColorPlot(SecretPlot):

    @property
    def main_assembler(self) -> Assembler:
        return ColorMapAssembler(self.fig, self._log)


class BooleanPlot(SecretPlot):

    def __init__(self, data, threshold):
        super().__init__(data)
        self.assembler.data.threshold = threshold

    @property
    def main_assembler(self) -> Assembler:
        return BooleanAssembler(self.fig, self._log)


def run():
    data = [6, 3, None, 2]
    h = (ColorPlot(data)
         .add_values()
         .add_y_padding(0, 1)
         .change_orientation("-x")
         .invert_y()
         .add_cmap("Accent", loc="top")
         .show()
         )



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 8:07 PM
#
# Color Manager

import matplotlib
from SecretColors import Palette, ColorMap
from SecretColors.utils import rgb_to_hex

from SecretPlots.constants import *
from SecretPlots.utils import Log


class ColorManager:
    def __init__(self, plot_type: str, log: Log):
        self._log = log
        self.plot_type = plot_type
        self._palette = None
        self._unique_colors = []
        self._all_colors = {}
        self._cmap = None
        self._on_color = None
        self._off_color = None
        self.user_colors = None
        self.user_cmap = None
        self._cycle = None

        self._log.info("ColorManager is initialized with default values.")

    @property
    def on_color(self):
        if self._on_color is None:
            self._on_color = self.palette.blue()
        return self._on_color

    @on_color.setter
    def on_color(self, value):
        self._on_color = value

    @property
    def off_color(self):
        if self._off_color is None:
            self._off_color = self.palette.red()
        return self._off_color

    @off_color.setter
    def off_color(self, value):
        self._off_color = value

    @property
    def palette(self) -> Palette:
        if self._palette is None:
            self._palette = Palette(show_warning=False)
        return self._palette

    @property
    def all_colors(self):
        return self._all_colors

    @property
    def cmap(self):
        if self.user_cmap is not None:
            self._cmap = matplotlib.cm.get_cmap(self.user_cmap)
        if self._cmap is None:
            colors = [self.palette.lime(shade=30), self.palette.lime(),
                      self.palette.brown(shade=40),
                      self.palette.brown(shade=80),
                      self.palette.black()]
            self._cmap = ColorMap(matplotlib, self.palette).from_list(colors)
        return self._cmap

    @staticmethod
    def cycle(color_list):
        while True:
            for c in color_list:
                yield c

    def _get_color(self, index):

        if self.user_colors is not None:
            if self._cycle is None:
                self._cycle = self.cycle(self.user_colors)
            if index not in self._all_colors.keys():
                self._all_colors[index] = next(self._cycle)

        if index not in self._all_colors.keys():
            if index < len(self.palette.get_color_list):
                self._all_colors[index] = self.palette.get_color_list[index]
            else:
                self._all_colors[index] = self.palette.random()

        return self._all_colors[index]

    def _get_from_map(self, value):
        r, g, b, _ = self.cmap(value)
        return rgb_to_hex(r, g, b)

    def color(self, position, value):
        row, col = position
        if self.plot_type == PLOT_BAR:
            if self.user_colors is not None:
                if len(self.user_colors) > 1:
                    return self._get_color(position[1])
                else:
                    return self._get_color(0)
            return self._get_color(row)
        if self.plot_type in [PLOT_STACKED_BAR, PLOT_GROUPED_BAR]:
            return self._get_color(col)
        elif self.plot_type == PLOT_COLOR_MAP:
            return self._get_from_map(value)
        elif self.plot_type == PLOT_BOOLEAN_PLOT:
            self._all_colors["ON"] = self.on_color
            self._all_colors["OFF"] = self.off_color
            if value == 0:
                return self.off_color
            else:
                return self.on_color
        else:
            return self._get_color(row)



# SecretPlots 2019
# Author : Rohit Suratekar
# Date : 3 September 2019
#
# All Classes and functions related to RNA-seq analysis

import matplotlib.pylab as plt
import pandas as pd
from SecretColors import Palette

from SecretPlots.constants.system import *
from SecretPlots.utils import log


class BaseAnalysis:
    def __init__(self, generate_log: bool = False,
                 logging_options: dict = None):
        self.generate_log = generate_log
        self.logging_options = logging_options
        self.palette = Palette()
        self.p = self.palette

    def analysis_name(self):
        raise NotImplementedError()

    def version_error(self):
        raise NotImplementedError()


class StringTie(BaseAnalysis):
    """
    Simple class for StringTie 2.0
    """

    # StringTie Columns

    COL_GENE_ID = "Gene ID"
    COL_GENE_NAME = "Gene Name"
    COL_REFERENCE = "Reference"
    COL_STRAND = "Strand"
    COL_START = "Start"
    COL_END = "End"
    COL_COVERAGE = "Coverage"
    COL_FPKM = "FPKM"
    COL_TPM = "TPM"

    def __init__(self, filename: str,
                 skip_header: bool = False,
                 generate_log: bool = False,
                 logging_options: dict = None):
        super().__init__(generate_log, logging_options)
        self.log("StringTie Object initiated", LOG_INFO)
        self.filename = filename
        self.skip_header = skip_header
        self.data = self.__parse_file()

    def version_error(self):
        raise Exception("This object accepts file generated by StringTie  "
                        "v2.0 ("
                        "http://ccb.jhu.edu/software/stringtie/index.shtml)."
                        " Please check the file format of this version.")

    def analysis_name(self):
        return "StringTie"

    def __parse_file(self):
        self.log("Attempting to parse {}".format(self.filename), LOG_INFO)
        with open(self.filename) as f:
            loc = f.tell()
            header = f.readline().split("\t")
            if len(header) != 9:
                self.log("Provided file is not in standard format", LOG_ERROR)
                raise self.version_error()
            if not self.skip_header:
                self.log("Header Skipped", LOG_INFO)
                f.seek(loc)

            return pd.read_csv(f, delimiter="\t")

    def log(self, message: str, log_type: int):
        if self.generate_log:
            if self.logging_options is None:
                log(message)
            else:
                self.logging_options["message"] = message
                log(**self.logging_options)

    @property
    def gene_ids(self):
        return self.data[StringTie.COL_GENE_ID].values

    @property
    def gene_names(self):
        return self.data[StringTie.COL_GENE_NAME].values

    @property
    def strands(self):
        return self.data[StringTie.COL_STRAND].values

    @property
    def locations(self):
        return self.data.apply(lambda x: (x[StringTie.COL_START],
                                          x[StringTie.COL_END]), axis=1).values

    @property
    def tpms(self):
        return self.data[StringTie.COL_TPM].values

    @property
    def fpkms(self):
        return self.data[StringTie.COL_FPKM].values

    @property
    def coverages(self):
        return self.data[StringTie.COL_COVERAGE].values

    @property
    def references(self):
        return self.data[StringTie.COL_REFERENCE].values

    @property
    def starts(self):
        return self.data[StringTie.COL_START].values

    @property
    def ends(self):
        return self.data[StringTie.COL_END].values

    @staticmethod
    def show():
        plt.show()

    def plot_tpms(self, min_value: float = 0, max_value: float = None):

        plt.hist(self.tpms)


def rr():
    s = StringTie('local/data.tsv')
    s.plot_tpms()
    s.show()


def run():
    rr()



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:57 PM
#
#
#
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 11:03 AM
#
#
# Main graph class

import matplotlib.pyplot as plt

from SecretPlots.assemblers import Assembler
from SecretPlots.utils import Log


class SecretPlot:
    def __init__(self, data, fig: plt.Figure = None, log: Log = None):
        if fig is None:
            fig = plt.figure()
        self.fig = fig
        self._raw_data = data
        if log is None:
            log = Log()
        self._log = log
        self._assembler = None
        self._figure_drawn = False

        self.orientation = None
        self.x_gap = None
        self.y_gap = None
        self.x_ticks = None
        self.y_ticks = None
        self.x_label = None
        self.y_label = None
        self.x_ticklabels = None
        self.y_ticklabels = None
        self.x_padding_start = None
        self.x_padding_end = None
        self.y_padding_start = None
        self.y_padding_end = None
        self.group_gap = None
        self.colors = None
        self.cmap = None
        self.cmap_location = None
        self.show_cmap = None
        self.show_legend = None
        self.width = None
        self.height = None
        self.shape = None
        self.show_x_midlines = None
        self.show_y_midlines = None
        self.show_x_edgelines = None
        self.show_y_edgelines = None
        self.show_values = None
        self.show_grid = None
        self.x_inverted = None
        self.y_inverted = None
        self.on_color = None
        self.off_color = None

        self._log.info(
            "'{}' initialization complete ".format(self.assembler.type))

    def _check_settings(self):
        if self.orientation is not None:
            self.assembler.am.orientation = self.orientation
        if self.x_gap is not None:
            self.x.gap = self.x_gap
        if self.y_gap is not None:
            self.y.gap = self.y_gap
        if self.group_gap is not None:
            self.assembler.am.group_gap = self.group_gap
        if self.colors is not None:
            if isinstance(self.colors, str):
                self.colors = [self.colors]
            self.assembler.cm.user_colors = self.colors
        if self.cmap is not None:
            self.assembler.cm.user_cmap = self.cmap

        self.assembler.gm.has_colorbar = self.show_cmap
        self.assembler.gm.colorbar_location = self.cmap_location
        self.assembler.em.show_legends = self.show_legend
        self.assembler.am.x.label = self.x_label
        self.assembler.am.y.label = self.y_label

        if self.width is not None:
            self.assembler.om.width = self.width
        if self.height is not None:
            self.assembler.om.height = self.height
        if self.shape is not None:
            self.assembler.om.shape = self.shape
        if self.show_values is not None:
            self.assembler.em.show_values = self.show_values
        if self.show_grid is not None:
            self.assembler.em.show_grid = self.show_grid

        if self.show_x_midlines is not None:
            self.assembler.am.x.show_midlines = self.show_x_midlines
        if self.show_y_midlines is not None:
            self.assembler.am.y.show_midlines = self.show_y_midlines
        if self.show_x_edgelines is not None:
            self.assembler.am.x.show_edgelines = self.show_x_edgelines
        if self.show_y_edgelines is not None:
            self.assembler.am.y.show_edgelines = self.show_y_edgelines

        if self.x_ticklabels is not None:
            self.assembler.am.x.tick_labels = self.x_ticklabels
        if self.y_ticklabels is not None:
            self.assembler.am.y.tick_labels = self.y_ticklabels
        if self.x_ticks is not None:
            self.assembler.am.x.ticks = self.x_ticks
        if self.y_ticks is not None:
            self.assembler.am.y.ticks = self.y_ticks

        if self.x_inverted is not None:
            self.assembler.am.x.is_inverted = self.x_inverted
        if self.y_inverted is not None:
            self.assembler.am.y.is_inverted = self.y_inverted

        if self.on_color is not None:
            self.assembler.cm.on_color = self.on_color
        if self.off_color is not None:
            self.assembler.cm.off_color = self.off_color

        if self.x_padding_start is not None:
            self.assembler.am.x.padding_start = self.x_padding_start
        if self.x_padding_end is not None:
            self.assembler.am.x.padding_end = self.x_padding_end
        if self.y_padding_start is not None:
            self.assembler.am.y.padding_start = self.y_padding_start
        if self.y_padding_end is not None:
            self.assembler.am.y.padding_end = self.y_padding_end

    @property
    def main_assembler(self) -> Assembler:
        raise NotImplementedError

    @property
    def x(self):
        return self.assembler.am.x

    @property
    def y(self):
        return self.assembler.am.y

    @property
    def ax(self) -> plt.Axes:
        return self.assembler.am.ax

    @property
    def assembler(self):
        if self._assembler is None:
            self._assembler = self.main_assembler
            self._assembler.data = self._raw_data
        return self._assembler

    def _assemble_components(self):
        self._check_settings()
        self.assembler.draw()

    def draw(self):
        if self._figure_drawn:
            return
        self._assemble_components()
        self._figure_drawn = True

    def show(self, tight=False):
        self.draw()
        if tight:
            plt.tight_layout()
        plt.show()

    def save(self, filename, **kwargs):
        self.draw()
        plt.savefig(filename, **kwargs)

    def add_grid(self, **kwargs):
        self.show_grid = True
        self.assembler.em.add_grid_options(**kwargs)
        return self

    def add_values(self, **kwargs):
        self.show_values = True
        self.assembler.em.add_value_options(**kwargs)
        return self

    def add_legends(self, **kwargs):
        self.show_legend = True
        self.assembler.em.add_legends_options(**kwargs)
        return self

    def add_x_ticklabels(self, labels, **kwargs):
        self.x_ticklabels = labels
        self.assembler.am.x.add_ticklabels_options(**kwargs)
        return self

    def add_y_ticklabels(self, labels, **kwargs):
        self.y_ticklabels = labels
        self.assembler.am.y.add_ticklabels_options(**kwargs)
        return self

    def add_x_gap(self, value):
        self.x_gap = value
        return self

    def add_y_gap(self, value):
        self.y_gap = value
        return self

    def add_x_midlines(self, **kwargs):
        self.show_x_midlines = True
        self.assembler.am.x.add_midlines_options(**kwargs)
        return self

    def add_y_midlines(self, **kwargs):
        self.show_y_midlines = True
        self.assembler.am.y.add_midlines_options(**kwargs)
        return self

    def add_x_edgelines(self, **kwargs):
        self.show_x_edgelines = True
        self.assembler.am.x.add_edgelines_options(**kwargs)
        return self

    def add_y_edgelines(self, **kwargs):
        self.show_y_edgelines = True
        self.assembler.am.y.add_edgelines_options(**kwargs)
        return self

    def add_group_gap(self, value):
        self.group_gap = value
        return self

    def invert_x(self):
        self.x_inverted = True
        return self

    def invert_y(self):
        self.y_inverted = True
        return self

    def add_x_label(self, label, **kwargs):
        self.x_label = label
        self.assembler.am.x.add_label_options(**kwargs)
        return self

    def add_y_label(self, label, **kwargs):
        self.y_label = label
        self.assembler.am.y.add_label_options(**kwargs)
        return self

    def add_x_padding(self, start, end):
        self.x_padding_start = start
        self.x_padding_end = end
        return self

    def add_y_padding(self, start, end):
        self.y_padding_start = start
        self.y_padding_end = end
        return self

    def add_cmap(self, name=None, loc="right"):
        self.show_cmap = True
        if name is not None:
            self.cmap = name

        self.cmap_location = loc
        return self

    def add_colors(self, values):
        self.colors = values
        return self

    def add_on_color(self, value):
        self.on_color = value
        return self

    def add_off_color(self, value):
        self.off_color = value
        return self

    def change_frame_visibility(self, left, right, top, bottom):
        self.assembler.am.frame_visibility = (left, right, top, bottom)
        return self

    def change_shape(self, value):
        self.shape = value
        return self

    def change_orientation(self, value):
        self.orientation = value
        return self

    def change_aspect_ratio(self, value):
        self.assembler.am.aspect_ratio = value
        return self

    def remove_cmap(self):
        self.show_cmap = False
        return self

    def remove_x_ticks(self):
        self.assembler.am.x.show_ticks = False
        return self

    def remove_y_ticks(self):
        self.assembler.am.y.show_ticks = False

    def remove_frame(self):
        self.assembler.am.frame_visibility = (0, 0, 0, 0)
        return self



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 11/13/19, 4:00 PM
#
#
#




#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:47 PM
#
#
#  Matrix Assemblers

from SecretPlots.assemblers import Assembler
from SecretPlots.constants import *
from SecretPlots.managers import ColorMapLocations


class ColorMapAssembler(Assembler):

    @property
    def main_location_manager(self):
        return ColorMapLocations(self.am, self.om, self._log)

    @property
    def type(self):
        return PLOT_COLOR_MAP

    def _adjust_defaults(self):

        if self.type != PLOT_BOOLEAN_PLOT:
            if self.gm.has_colorbar is None:
                self.gm.has_colorbar = True
            if self.gm.colorbar_location is None:
                self.gm.colorbar_location = "right"
            else:
                if self.gm.has_colorbar is None:
                    self.gm.has_colorbar = True

        else:
            self.em.show_legends = True

    def _draw_elements(self):
        locations = self.lm.get(self.data)
        for loc, val, pos in zip(locations, self.data.value,
                                 self.data.positions):
            x, y = loc
            if self.am.orientation == "y":
                x, y = y, x

            if self.type == PLOT_BOOLEAN_PLOT:
                if self.data.threshold is None:
                    self._log.error("For BooleanPlot, you should specify "
                                    "threshold")
                v = 1 if val >= self.data.threshold else 0
                shape = self.om.get(x, y, v, pos)
            else:
                shape = self.om.get(x, y, val / self.data.max, pos)

            self.ax.add_patch(shape.get())
            self.em.draw_values(shape, val,
                                self.cm.color(pos, val / self.data.max))

        ticks_major = []
        ticks_minor = []
        edge_major = []
        edge_minor = []
        for i, m in enumerate(locations):
            if i == 0:
                edge_major.append(m[0])
                edge_minor.append(m[1])
            elif i == len(locations) - 1:
                edge_major.append(m[0] + self.om.width)
                edge_minor.append(m[1] + self.om.height)
            temp = round(m[0] + self.om.width / 2, 2)
            temp_minor = round(m[1] + self.om.height / 2, 2)
            if temp not in ticks_major:
                ticks_major.append(temp)
            if temp_minor not in ticks_minor:
                ticks_minor.append(temp_minor)

        self.am.major.make_ticks(ticks_major)
        self.am.minor.make_ticks(ticks_minor)
        self.am.major.edgelines = edge_major
        self.am.minor.edgelines = edge_minor

        self.am.major.make_labels()
        self.am.minor.make_labels()

    def draw(self):
        self._adjust_defaults()
        self._draw_elements()
        self._draw_axis()
        self._draw_extra()


class BooleanAssembler(ColorMapAssembler):

    @property
    def type(self):
        return PLOT_BOOLEAN_PLOT



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:52 PM
#
# All Location Managers will go here

from SecretPlots.managers._axis import AxisManager
from SecretPlots.managers._object import ObjectManager
from SecretPlots.objects import Data
from SecretPlots.utils import Log


class LocationManager:
    def __init__(self, am: AxisManager, om: ObjectManager, log: Log):
        self._log = log
        self._major = None
        self._minor = None
        self.om = om
        self.start_point = (0, 0)
        self.am = am

    @property
    def width(self):
        return self.om.width

    @property
    def height(self):
        return self.om.height

    @property
    def major_gap(self):
        return self.am.major.gap

    @property
    def minor_gap(self):
        return self.am.minor.gap

    @property
    def plot_type(self):
        raise NotImplementedError

    @property
    def major(self):
        if self._major is None:
            self._major, self._minor = self.start_point
        return self._major

    @property
    def minor(self):
        if self._minor is None:
            self._major, self._minor = self.start_point
        return self._minor

    def validate(self, data: Data):
        raise NotImplementedError

    def get(self, data: Data):
        raise NotImplementedError



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 8:03 PM
#
#
# Grid Manager

import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

from SecretPlots.utils import Log


class GridManager:
    def __init__(self, fig: plt.Figure, log: Log):
        self.fig = fig
        self._log = log
        self._log.info("GridManager is initialized with default values")
        self._main = None
        self._cb = None
        self.has_colorbar = None
        self._cb_location = None
        self._ax_grid = None

    @property
    def colorbar_location(self):
        if self._cb_location is None:
            self._cb_location = "right"
        return self._cb_location

    @colorbar_location.setter
    def colorbar_location(self, value):
        self._cb_location = value

    @property
    def ax_grid(self):
        if self._ax_grid is None:
            self._log.info("Default GridSpec with 3 rows and 3 columns is "
                           "generated")
            self._ax_grid = GridSpec(3, 3,
                                     width_ratios=[1, 15, 1],
                                     height_ratios=[1, 15, 1],
                                     figure=self.fig)
        return self._ax_grid

    def _generate_axes(self):
        if not self.has_colorbar:
            self._main = self.fig.add_subplot(111)
            self._log.info("Plot Grid is set to normal.")
            return

        if self._cb_location in ["right", "r"]:
            main = self.ax_grid[:, :-1]
            cb = self.ax_grid[:, -1]
        elif self._cb_location in ["left", "l"]:
            main = self.ax_grid[:, 1:]
            cb = self.ax_grid[:, 0]
        elif self._cb_location in ["top", "t"]:
            main = self.ax_grid[1:, :]
            cb = self.ax_grid[0, :]
        elif self._cb_location in ["bottom", "b"]:
            main = self.ax_grid[:-1, :]
            cb = self.ax_grid[-1, :]
        else:
            self._log.error("No such colorbar location found : {}".format(
                self._cb_location))

        self._main = self.fig.add_subplot(main)
        self._cb = self.fig.add_subplot(cb)
        self._log.info("Plot Grid is set according to colorbar location")

    def get_main_axis(self) -> plt.Axes:
        if self._main is None:
            self._generate_axes()
        return self._main

    def get_colorbar_axis(self) -> plt.Axes:
        if self._main is None:
            self._generate_axes()
        return self._cb



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 8:04 PM
#
#
# Axis Manager

from SecretPlots.objects import Axis
from SecretPlots.utils import Log


class AxisManager:
    def __init__(self, gm, log: Log):
        self.gm = gm
        self._log = log

        self._orientation = None
        self._x = None
        self._y = None
        self._ax = None
        self.is_reversed = False
        self.aspect_ratio = None
        self.group_gap = 1
        self._frame_visibility = None

        self._log.info("AxisManager is initialized with default values")

    @property
    def frame_visibility(self):
        if self._frame_visibility is None:
            self._frame_visibility = (1, 1, 1, 1)
        return self._frame_visibility

    @frame_visibility.setter
    def frame_visibility(self, value):
        if type(value) == bool:
            if value:
                self._frame_visibility = (1, 1, 1, 1)
            else:
                self._frame_visibility = (0, 0, 0, 0)
        elif type(value) == tuple:
            if len(value) == 4:
                self._frame_visibility = value
            else:
                self._log.error("Frame visibility should be either Bool or "
                                "tuple with length 4 (left, right, top , "
                                "bottom)")

    @property
    def orientation(self):
        if self._orientation is None:
            self._orientation = "x"
        return self._orientation.replace("-", "").replace("+", "")

    @orientation.setter
    def orientation(self, value: str):
        value = value.strip().lower()
        if value.replace("-", "").replace("+", "") not in ["x", "y"]:
            self._log.error("Invalid axis orientation. Allowed orientations "
                            "are x, y, -x, -y")

        self._orientation = value
        self._log.info("Axis orientation set to : {}".format(value))

        if self._orientation == "-x":
            self.y.is_inverted = True
        elif self._orientation == "-y":
            self.x.is_inverted = True

    @property
    def is_reflected(self):
        return "-" in self.orientation

    @property
    def ax(self):
        if self._ax is None:
            self._ax = self.gm.get_main_axis()
        return self._ax

    @property
    def x(self):
        if self._x is None:
            self._x = Axis("x", 0, self._log)
        return self._x

    @property
    def y(self):
        if self._y is None:
            self._y = Axis("y", 1, self._log)
        return self._y

    @property
    def major(self):
        if self.orientation == "x":
            return self.x
        else:
            return self.y

    @property
    def minor(self):
        if self.orientation == "x":
            return self.y
        else:
            return self.x



# SecretPlots 2019
# Author : Rohit Suratekar
# Date : 4 September 2019
#
# Simple utils to use in the library

import logging


class Log:
    def __init__(self, show_log: bool = False,
                 options: dict = None,
                 add_to_console: bool = True,
                 add_to_file: bool = False,
                 filename: str = "script.log",
                 logging_format: str = "%(asctime)s %(filename)s : %(message)s"):
        self.show_log = show_log
        self.formatter = logging.Formatter(logging_format)
        self.add_to_console = add_to_console
        self.add_to_file = add_to_file
        self.filename = filename
        self._log_object = None

    @property
    def log_object(self):
        if self._log_object is None:
            self._log_object = logging.getLogger("log")
            if self.add_to_console:
                console = logging.StreamHandler()
                console.setFormatter(self.formatter)
                self._log_object.addHandler(console)

            if self.add_to_file:
                log_file = logging.FileHandler(self.filename)
                log_file.setFormatter(self.formatter)
                self._log_object.addHandler(log_file)
        return self._log_object

    def info(self, message):
        if self.show_log:
            self.log_object.setLevel(logging.INFO)
            self.log_object.info(message)

    def error(self, message, raise_exception=True):
        if self.show_log:
            self.log_object.setLevel(logging.ERROR)
            self.log_object.error(message)

        if raise_exception:
            raise Exception(message)

    def warn(self, message):
        if self.show_log:
            self.log_object.setLevel(logging.WARN)
            self.log_object.warning(message)


def run():
    pass



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 03/10/19, 3:57 PM
#
#
# All graph managers

import matplotlib.pyplot as plt

from SecretPlots.managers import *
from SecretPlots.objects import Data
from SecretPlots.utils import Log


class Assembler:

    def __init__(self, fig: plt.Figure, log: Log):
        self._log = log
        self._fig = fig

        self._type = None
        self._data = None
        self._am = None
        self._lm = None
        self._gm = None
        self._om = None
        self._cm = None
        self._em = None

        self._log.info("Assembler is initialized with default values")

    @property
    def gm(self) -> GridManager:
        if self._gm is None:
            self._gm = GridManager(self._fig, self._log)
        return self._gm

    @property
    def cm(self) -> ColorManager:
        if self._cm is None:
            self._cm = ColorManager(self.type, self._log)
        return self._cm

    @property
    def am(self) -> AxisManager:
        if self._am is None:
            self._am = AxisManager(self.gm, self._log)
        return self._am

    @property
    def om(self) -> ObjectManager:
        if self._om is None:
            self._om = ObjectManager(self.am, self.cm, self._log)
        return self._om

    @property
    def em(self) -> ExtraManager:
        if self._em is None:
            self._em = ExtraManager(self.gm, self.am, self.cm, self._log)
        return self._em

    @property
    def main_location_manager(self) -> LocationManager:
        raise NotImplementedError

    @property
    def lm(self) -> LocationManager:
        if self._lm is None:
            self._lm = self.main_location_manager
        return self._lm

    @property
    def ax(self) -> plt.Axes:
        return self.am.ax

    @property
    def type(self):
        raise NotImplementedError

    def draw(self):
        raise NotImplementedError

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def data(self) -> Data:
        return self._data

    @data.setter
    def data(self, value):
        self._data = Data(value, self._log)

    def _set_auto_limit(self):

        if self.am.x.padding_start is None:
            self.am.x.padding_start = 1
        if self.am.y.padding_start is None:
            self.am.y.padding_start = 1
        if self.am.x.padding_end is None:
            self.am.x.padding_end = 1
        if self.am.y.padding_end is None:
            self.am.y.padding_end = 1

        self.ax.set_xlim(self.om.min_x - self.am.x.padding_start,
                         self.om.max_x + self.am.x.padding_end)

        self.ax.set_ylim(self.om.min_y - self.am.y.padding_start,
                         self.om.max_y + self.am.y.padding_end)

        self._log.info("Axis limit is set automatically")

    def _check_axis_transformations(self):
        if self.am.x.is_inverted:
            self.ax.invert_xaxis()
        if self.am.y.is_inverted:
            self.ax.invert_yaxis()

        if self.am.x.scale is not None:
            self.ax.set_xscale(self.am.x.scale)
        if self.am.y.scale is not None:
            self.ax.set_yscale(self.am.y.scale)

    def _draw_axis(self):
        self._set_auto_limit()
        if len(self.am.x.ticks) != 0:
            self.ax.set_xticks(self.am.x.ticks, **self.am.x.tick_options)
            self.ax.set_xticklabels(self.am.x.tick_labels,
                                    **self.am.x.ticklabels_options)
        if len(self.am.y.ticks) != 0:
            self.ax.set_yticks(self.am.y.ticks, **self.am.y.tick_options)
            self.ax.set_yticklabels(self.am.y.tick_labels,
                                    **self.am.y.ticklabels_options)

        left, right, top, bottom = self.am.frame_visibility
        self.ax.spines['left'].set_visible(left)
        self.ax.spines['right'].set_visible(right)
        self.ax.spines['top'].set_visible(top)
        self.ax.spines['bottom'].set_visible(bottom)

        if self.am.aspect_ratio is not None:
            self.ax.set_aspect(self.am.aspect_ratio)

        if not self.am.x.show_ticks:
            self.ax.set_xticks([])
        if not self.am.y.show_ticks:
            self.ax.set_yticks([])

        if self.am.x.label is not None:
            self.ax.set_xlabel(self.am.x.label, **self.am.x.label_options)
        if self.am.y.label is not None:
            self.ax.set_ylabel(self.am.y.label, **self.am.y.label_options)

        self._check_axis_transformations()

    def _draw_extra(self):
        self.am.major.make_midlines()
        self.am.minor.make_midlines()

        self.em.draw_midlines()
        self.em.draw_edgelines()
        self.em.draw_colorbar(self.data)
        self.em.draw_legends()
        self.em.draw_grid()
        if self.om.show_missing:
            self.em.draw_missing(**self.om.missing_options)


def run():
    data = [[3, 4, 1], [1, 3], 2]



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 11/15/19, 5:10 PM
#
#
# Constant related to the network

# Directions

RIGHT = 0
BOTTOM_RIGHT = 1
BOTTOM = 2
BOTTOM_LEFT = 3
LEFT = 4
TOP_LEFT = 5
TOP = 6
TOP_RIGHT = 7

DIRECTIONS = [
    RIGHT,
    BOTTOM_RIGHT,
    BOTTOM,
    BOTTOM_LEFT,
    LEFT,
    TOP_LEFT,
    TOP,
    TOP_RIGHT
]

DIRECTION_TEXT = [
    "RIGHT",
    "BOTTOM_RIGHT",
    "BOTTOM",
    "BOTTOM_LEFT",
    "LEFT",
    "TOP_LEFT",
    "TOP",
    "TOP_RIGHT"
]



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:53 PM
#
#
# Matrix Locators

from SecretPlots.constants.graph import *
from SecretPlots.managers.location._base import LocationManager
from SecretPlots.objects import Data


class ColorMapLocations(LocationManager):

    @property
    def plot_type(self):
        return PLOT_COLOR_MAP

    def validate(self, data: Data):
        self._log.info("Data validated fo ColorMap")
        return True

    def _single_column(self, data: Data):

        self._log.info("Calculating positions for simple Bars")
        points = []
        for loc in data.positions:
            points.append((
                self.major,
                self.minor + loc[1] * (self.height + self.minor_gap)
            ))
        return points

    def _matrix_columns(self, data: Data):
        points = []
        for loc in data.positions:
            row, col = loc
            points.append((
                self.major + row * (self.width + self.major_gap),
                self.minor + col * (self.height + self.minor_gap)
            ))
        return points

    def _complex_columns(self, data: Data):
        points = []
        stack = None
        col_no = 0
        for loc in data.positions:
            row, col = loc
            if stack is None:
                stack = self.minor
            if col_no != row:
                stack = self.minor
                col_no = row
            points.append((
                self.major + row * (self.width + self.major_gap),
                stack
            ))
            stack += self.height + self.minor_gap

        return points

    def get(self, data: Data):
        self.validate(data)
        if data.type in [Data.SINGLE_VALUED, Data.SIMPLE_CATEGORICAL]:
            return self._single_column(data)
        elif data.type in [Data.POINTS, Data.MATRIX]:
            return self._matrix_columns(data)
        elif data.type == Data.COMPLEX_CATEGORICAL:
            return self._complex_columns(data)
        else:
            self._log.error("Data type {} does not support ColorMap".format(
                data.type_name))



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:50 PM
#

from SecretPlots.managers._axis import AxisManager
from SecretPlots.managers._color import ColorManager
from SecretPlots.managers._extra import ExtraManager
from SecretPlots.managers._grid import GridManager
from SecretPlots.managers._object import ObjectManager

from SecretPlots.managers.location import *



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:51 PM
#
#

from SecretPlots.managers.location._base import LocationManager

from SecretPlots.managers.location._bars import (BarLocations,
                                                 BarGroupLocations,
                                                 HistLocations)

from SecretPlots.managers.location._matrix import ColorMapLocations



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:55 PM
#
#
#
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 30/09/19, 10:11 AM
#
#
# Extra Manager

import matplotlib
import matplotlib.pyplot as plt
from SecretColors.utils import text_color
from matplotlib.patches import Patch

from SecretPlots.managers._axis import AxisManager
from SecretPlots.managers._color import ColorManager
from SecretPlots.managers._grid import GridManager
from SecretPlots.objects import Data
from SecretPlots.utils import Log


class ExtraManager:
    def __init__(self, gm: GridManager, am: AxisManager, cm: ColorManager,
                 log: Log):
        self.gm = gm
        self.cm = cm
        self.am = am
        self._log = log
        self.show_values = False
        self.show_legends = False
        self.show_grid = False
        self._value_options = None
        self._legends_options = None
        self._grid_options = None

        self._log.info("ExtraManager is initialized with default values")

    @property
    def value_options(self):
        if self._value_options is None:
            self._value_options = {
                "anchor": (0.5, 0.5),
                "relative": True,
                "offset": (0, 0)
            }
        return self._value_options

    @property
    def grid_options(self):
        if self._grid_options is None:
            self._grid_options = {}
        return self._grid_options

    @property
    def legends_options(self):
        if self._legends_options is None:
            self._legends_options = {}
        return self._legends_options

    def add_value_options(self, **kwargs):
        self._value_options = {**self.value_options, **kwargs}

    def add_legends_options(self, **kwargs):
        self._legends_options = {**self.legends_options, **kwargs}

    def add_grid_options(self, **kwargs):
        self._grid_options = {**self.grid_options, **kwargs}

    def draw_values(self, shape, text, bg_color):
        if not self.show_values:
            return

        opts = {"ha": "center", "color": text_color(bg_color)}

        opts = {**opts, **self.value_options}

        if opts["relative"]:
            x = shape.x + shape.width * opts["anchor"][0] + opts["offset"][0]
            y = shape.y + shape.height * opts["anchor"][1] + opts["offset"][1]
        else:
            x = shape.x + opts["anchor"][0] + opts["offset"][0]
            y = shape.y + opts["anchor"][1] + opts["offset"][1]

        if opts["anchor"][0] != 0.5 and opts["anchor"][1] != 0.5:
            del opts["ha"]

        del opts["anchor"]
        del opts["relative"]
        del opts["offset"]

        self.gm.get_main_axis().text(x, y, "{}".format(text), **opts)
        self._log.info("Values are shown in the plot")

    def draw_midlines(self):
        if self.am.x.show_midlines:
            for x in self.am.x.midlines:
                self.am.ax.axvline(x, **self.am.x.midlines_options)

        if self.am.y.show_midlines:
            for y in self.am.y.midlines:
                self.am.ax.axhline(y, **self.am.y.midlines_options)

        self._log.info("Midlines are added to the plot")

    def draw_edgelines(self):
        if self.am.x.show_edgelines:
            for x in self.am.x.edgelines:
                self.am.ax.axvline(x, **self.am.x.edgelines_options)

        if self.am.y.show_edgelines:
            for y in self.am.y.edgelines:
                self.am.ax.axhline(y, **self.am.y.edgelines_options)

        self._log.info("Edgelines are added the plot")

    def draw_colorbar(self, data: Data):

        if not self.gm.has_colorbar:
            return

        norm = matplotlib.colors.Normalize(vmin=data.min, vmax=data.max)
        sm = plt.cm.ScalarMappable(cmap=self.cm.cmap, norm=norm)

        if self.gm.colorbar_location in ["right", "left", "r", "l"]:
            ori = "vertical"
        else:
            ori = "horizontal"

        plt.colorbar(sm, cax=self.gm.get_colorbar_axis(),
                     orientation=ori)

        if self.gm.colorbar_location in ["left", "l"]:
            self.gm.get_colorbar_axis().yaxis.set_ticks_position('left')
        elif self.gm.colorbar_location in ["top", "t"]:
            self.gm.get_colorbar_axis().xaxis.set_ticks_position("top")

        self._log.info("Colorbar is added the plot")

    def draw_legends(self):
        if not self.show_legends:
            return

        if len(self.cm.all_colors) == 0:
            self._log.warn("No legends were created because no color patches "
                           "found. Most likely, your plot is using colormap "
                           "for getting colors. Use 'show_cmap' instead. ")

        p = []
        for key in self.cm.all_colors:
            p.append(Patch(
                facecolor=self.cm.all_colors[key],
                label="{}".format(key)
            ))

        opts = {
            "handles": p
        }
        opts = {**opts, **self.legends_options}
        plt.legend(**opts)
        self._log.info("Legends are added to the plot")

    def draw_grid(self):
        if not self.show_grid:
            return
        plt.grid(**self.grid_options)
        self._log.info("Plot grid is generated")

    def draw_missing(self, **kwargs):
        for x in self.am.x.missing_regions:
            plt.axvspan(x[0], x[1], **kwargs)



import setuptools

with open("PYPI.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SecretPlots",
    version="0.1.1",
    author="Rohit Suratekar",
    author_email="rohitsuratekar@gmail.com",
    description="Make plotting great again!",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/secretBiology/SecretPlots",
    packages=setuptools.find_packages(),
    license='MIT License',
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 1 - Planning",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
)



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 8:09 PM
#
# Object Manager

import numpy as np

from SecretPlots.constants import *
from SecretPlots.managers._axis import AxisManager
from SecretPlots.managers._color import ColorManager
from SecretPlots.objects import Element
from SecretPlots.utils import Log


class ObjectManager:
    def __init__(self, am: AxisManager, cm: ColorManager, log: Log):
        self._log = log
        self.am = am
        self.cm = cm
        self.is_point = False
        self.width = 1
        self.height = 1
        self.rotation = 0
        self._options = None
        self._missing_options = None
        self.shape = "r"
        self.max_x = 0
        self.max_y = 0
        self.min_x = 0
        self.min_y = 0
        self.show_missing = True
        self._no_of_missing = 0
        self._log.info("ObjectManager is initialized with default values")

    @property
    def options(self):
        if self._options is None:
            self._options = {}
        return self._options

    @property
    def missing_options(self):
        if self._missing_options is None:
            self._missing_options = {
                "fill": False,
                "hatch": "//",
                "zorder": 0
            }
        return self._missing_options

    def add_options(self, **kwargs):
        self._options = {**self.options, **kwargs}

    def add_missing_options(self, **kwargs):
        self._missing_options = {**self.missing_options, **kwargs}

    def _check_limits(self, x, y):
        if x > self.max_x:
            self.max_x = x
        elif x < self.min_x:
            self.min_x = x
        if y > self.max_y:
            self.max_y = y
        elif y < self.min_y:
            self.min_y = y

    def _get_bar_object(self, x, y, value, pos):
        e = Element(self._log)
        e.add_options(**self.options)

        if np.isnan(value):
            self._no_of_missing += 1
            value = 0
        else:
            e.add_options(color=self.cm.color(pos, value))
        e.shape = self.shape
        if self.am.orientation == "x":
            e.width = self.width
            e.height = value
        else:
            e.width = value
            e.height = self.width

        self._check_limits(x, y)
        self._check_limits(x + e.width, y + e.height)
        return e.get(x, y)

    def _get_network_object(self, x, y, value, pos):
        e = Element(self._log)
        e.add_options(**self.options)
        e.width = self.width
        e.height = self.height
        e.add_options(color=self.cm.color(pos, value))
        self._check_limits(x - e.width / 2, y - e.height / 2)
        self._check_limits(x + e.width / 2, y + e.height / 2)
        return e.get(x - e.width / 2, y - e.height / 2)

    def _get_colormap_object(self, x, y, value, pos):
        e = Element(self._log)
        if np.isnan(value):
            self._no_of_missing += 1
            e.add_options(**self.missing_options)
        else:
            e.add_options(**self.options)
        e.add_options(color=self.cm.color(pos, value))
        e.shape = self.shape
        e.width = self.width
        e.height = self.height
        self._check_limits(x, y)
        self._check_limits(x + e.width, y + e.height)
        return e.get(x, y)

    def get(self, x, y, value, pos):
        if self.cm.plot_type in [PLOT_BAR, PLOT_STACKED_BAR, PLOT_GROUPED_BAR]:
            return self._get_bar_object(x, y, value, pos)
        elif self.cm.plot_type in [PLOT_COLOR_MAP, PLOT_BOOLEAN_PLOT]:
            return self._get_colormap_object(x, y, value, pos)
        elif self.cm.plot_type in [PLOT_NETWORK]:
            return self._get_network_object(x, y, value, pos)
        else:
            self._log.error("ObjectManager for {} is not set".format(
                self.cm.plot_type))



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:52 PM
#
# Bar Locations

import numpy as np

from SecretPlots.constants import *
from SecretPlots.managers._axis import AxisManager
from SecretPlots.managers.location._base import LocationManager
from SecretPlots.objects import Data
from SecretPlots.utils import Log


class BarLocations(LocationManager):

    @property
    def plot_type(self):
        return PLOT_BAR

    def validate(self, data: Data):
        self._log.info("Valid data is provided for the BarPlot")
        return True

    def _simple_bars(self, data: Data):
        self._log.info("Calculating positions for simple Bars")
        points = []
        for loc in data.positions:
            points.append((
                self.major + loc[1] * (self.width + self.major_gap),
                self.minor
            ))
        return points

    def _stacked_bars(self, data: Data):
        self._log.info("Calculating positions for Stacked Bars")
        points = []
        stack = None
        last_col = 0
        for loc, value in zip(data.positions, data.value):
            if np.isnan(value):
                value = 0
                self._log.warn("NaN value found, ignoring its effect")
            m1, m2 = loc
            if stack is None:
                stack = self.minor
            if m1 != last_col:
                stack = self.minor
                last_col += 1

            points.append((
                self.major + m1 * (self.width + self.major_gap),
                stack
            ))
            stack += value + self.minor_gap
        return points

    def get(self, data: Data) -> list:
        self.validate(data)

        if data.type in [Data.SINGLE_VALUED, Data.SIMPLE_CATEGORICAL]:
            return self._simple_bars(data)
        elif data.type in [Data.COMPLEX_CATEGORICAL, Data.MATRIX]:
            return self._stacked_bars(data)
        elif data.type == Data.POINTS:
            if data.is_single_point:
                return self._simple_bars(data)
            else:
                return self._stacked_bars(data)
        else:
            self._log.error("This data type is nor supported for BarPlot")


class BarGroupLocations(LocationManager):

    def __init__(self, am: AxisManager, om, log: Log):
        super().__init__(am, om, log)
        self.group_gap = am.group_gap

    @property
    def plot_type(self):
        return PLOT_GROUPED_BAR

    def validate(self, data: Data):
        self._log.info("Data validated fo GroupedBarPlot")
        return True

    def _simple_bars(self, data: Data):
        self._log.info("Calculating positions for simple Bars")
        points = []
        for loc in data.positions:
            points.append((
                self.major + loc[1] * (self.width + self.major_gap),
                self.minor
            ))
        return points

    def _grouped_bars(self, data: Data):
        self._log.info("Calculating positions for Grouped Bars")
        points = []
        bars = -1
        for loc, value in zip(data.positions, data.value):
            m1, m2 = loc
            bars += 1
            points.append((
                self.major + bars * (self.width + self.major_gap) +
                m1 * self.group_gap,
                self.minor
            ))
        return points

    def get(self, data: Data) -> list:
        self.validate(data)

        if data.type in [Data.SINGLE_VALUED, Data.SIMPLE_CATEGORICAL]:
            return self._simple_bars(data)
        elif data.type in [Data.COMPLEX_CATEGORICAL, Data.MATRIX]:
            return self._grouped_bars(data)
        elif data.type == Data.POINTS:
            if data.is_single_point:
                return self._simple_bars(data)
            else:
                return self._grouped_bars(data)
        else:
            self._log.error("This data type is nor supported for "
                            "GroupedBarPlot")


class HistLocations(LocationManager):

    @property
    def plot_type(self):
        return PLOT_HIST

    def __init__(self, am: AxisManager, om, log: Log, bins=None):
        super().__init__(am, om, log)
        if bins is None:
            bins = "autp"
        self.bins = bins
        self._hist_options = {}

    def validate(self, data: Data):
        if data.type != Data.SIMPLE_CATEGORICAL:
            self._log.warn("Data will be flatten for histogram")
        self._log.info("Valid data is provided for Histogram")

    def get(self, data: Data):
        self.validate(data)

        bins, _ = np.histogram(data.value, self.bins, **self._hist_options)
        return [(self.major + x * (self.width + self.major_gap), self.minor)
                for x in range(len(bins))]

    def hist_options(self, **kwargs):
        self._hist_options = {**self._hist_options, **kwargs}



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 03/10/19, 3:20 PM
#
#
# Graph constants

PLOT_BAR = "bar"
PLOT_HIST = "hist"
PLOT_COLOR_MAP = "colormap"
PLOT_STACKED_BAR = "stacked_bar"
PLOT_GROUPED_BAR = "grouped_bar"
PLOT_BOOLEAN_PLOT = "boolean"
PLOT_NETWORK = "network"



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 05/10/19, 7:45 PM
#
#
# Bar Assemblers

import numpy as np

from SecretPlots.assemblers import Assembler
from SecretPlots.constants.graph import *
from SecretPlots.managers import (BarLocations,
                                  BarGroupLocations,
                                  LocationManager)
from SecretPlots.objects import Data


class BarAssembler(Assembler):

    @property
    def type(self):
        if self.data.type in [Data.SINGLE_VALUED, Data.SIMPLE_CATEGORICAL]:
            return PLOT_BAR
        else:
            if self.data.is_single_point:
                return PLOT_BAR
            return PLOT_STACKED_BAR

    @property
    def main_location_manager(self) -> LocationManager:
        return BarLocations(self.am, self.om, self._log)

    def _adjust_defaults(self):
        if not self.am.major.is_user_defined_gap:
            self.am.major.gap = 0.2

        self.am.minor.padding_start = self.am.minor.padding_start or 0

        if self.type == PLOT_STACKED_BAR:
            if self.em.show_legends is None:
                self.em.show_legends = True

    def _draw_elements(self):
        locations = self.lm.get(self.data)
        for loc, val, pos in zip(locations, self.data.value,
                                 self.data.positions):
            x, y = loc
            if self.am.orientation == "y":
                x, y = y, x

            shape = self.om.get(x, y, val, pos)
            if not np.isnan(val) or self.om.show_missing:
                if np.isnan(val):
                    self.am.major.add_missing_region(
                        [x, y][self.am.major.index], self.om.width)
                else:
                    self.ax.add_patch(shape.get())
            self.em.draw_values(shape, val, self.cm.color(pos, val))

        ticks = []
        edge = []
        for i, m in enumerate(locations):
            if i == 0:
                edge.append(m[0])
            elif i == len(locations) - 1:
                edge.append(m[0] + self.om.width)
            temp = round(m[0] + self.om.width / 2, 2)
            if temp not in ticks:
                ticks.append(temp)

        self.am.major.make_ticks(ticks)
        self.am.major.edgelines = edge
        self.am.major.make_labels()

    def draw(self):
        self._adjust_defaults()
        self._draw_elements()
        self._draw_axis()
        self._draw_extra()


class BarGroupedAssembler(Assembler):
    @property
    def type(self):
        return PLOT_GROUPED_BAR

    @property
    def main_location_manager(self) -> LocationManager:
        return BarGroupLocations(self.am, self.om, self._log)

    def _adjust_defaults(self):

        self.am.minor.padding_start = self.am.minor.padding_start or 0

        if self.em.show_legends is None:
            self.em.show_legends = True

    def _draw_elements(self):
        locations = self.lm.get(self.data)
        for loc, val, pos in zip(locations, self.data.value,
                                 self.data.positions):
            x, y = loc
            if self.am.orientation == "y":
                x, y = y, x

            shape = self.om.get(x, y, val, pos)
            if not np.isnan(val) or self.om.show_missing:
                if np.isnan(val):
                    self.am.major.add_missing_region(
                        [x, y][self.am.major.index], self.om.width)
                else:
                    self.ax.add_patch(shape.get())
            self.em.draw_values(shape, val, self.cm.color(pos, val))

        ticks = []
        edge = []
        current_group = 0
        last_item = None
        for pos, loc in zip(self.data.positions, locations):
            if current_group == pos[0]:
                current_group += 1
                if last_item is not None:
                    edge.append(last_item)
                edge.append(loc)
            last_item = loc
        edge.append(last_item)
        edge = [x[0] for x in edge]
        midlines = []
        for e in range(len(edge)):
            if e % 2 == 1:
                edge[e] += self.om.width
                ticks.append((edge[e - 1] + edge[e]) / 2)
                try:
                    midlines.append((edge[e] + edge[e + 1]) / 2)
                except IndexError:
                    pass

        self.am.major.make_ticks(ticks)
        self.am.major.edgelines = edge
        self.am.major.midlines = midlines

        self.am.major.make_labels()

    def draw(self):
        self._adjust_defaults()
        self._draw_elements()
        self._draw_axis()
        self._draw_extra()



#  SecretPlots
#  Copyright (c) 2019.  SecretBiology
#
#  Author: Rohit Suratekar
#  Organisation: SecretBiology
#  Website: https://github.com/secretBiology/SecretPlots
#  Licence: MIT License
#  Creation: 11/16/19, 10:18 PM
#
#
# Tests for network module

from SecretPlots.utils import Log
from SecretPlots.network.pathfinder import PathFinder, Space
import numpy as np
from SecretPlots.constants.network import *
import pytest


def get_data():
    data = [[0, 0, 0, 0],
            [0, None, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0]]
    return data


def test_pathfinder():
    p = PathFinder(get_data())


def test_space():
    data = [
        ["a", "b", 1],
        ["a", "c", 1],
        ["a", "d", 1],
        ["b", "c", 1],
    ]

    s = Space(data)

    # Check node placement
    assert s.max_cols == 2
    assert len(s.nodes) == 4
    with pytest.raises(ValueError):
        s.max_cols = 1.5
